#!/usr/bin/awk -f
#
#  MIB to HTML file converter
#  Copyright (C) 1996 by Equivalance
#
#  $Log: mib2html.awk,v $
#  Revision 1.3  1996/08/11 12:44:56  craigs
#  Added support for dateandtime
#  Fixed problem with truncating sequences
#
#  Revision 1.2  1996/08/11 05:43:37  craigs
#  Added lots of extra info
#
#  Revision 1.1  1996/08/10 13:54:50  craigs
#  Initial revision
#
#  Revision 1.1  1996/08/03 11:57:04  craigs
#  Initial revision
#
#

#
#  This script will convert most MIB definition files into a tree of
#  HTML files that can be used by MibMaster. It requires the "awk"
#  program, and is invoked as follows:
#
#          awk -f mib_to_html mib.txt
#
#      where:   mib.txt   is the input ASN file
#
#  This script will generate a single HTML file, and a directory containing
#  any secondary HTML files needed. Simply move these to the "mibs" directory
#  and MibMaster will recognise the file as a Custom MIB file. You will have
#  to reset MibMaster before the Custom MIB file list is updated. This can
#  be done by going to the Global Configuration page and pressing the Accept
#  button without changing any paramaters.
#
#  This script will correctly interpret tables, and will automatically
#  detect and fill in enumerated types. It extracts the MIB descriptions
#  and inserts them as online help automatically. In fact, it's pretty damn
#  good, and will take every MIB file I can throw at it!!
#
#  Although other versions of AWK may work, I recommend the GNU AWK. 
#  A DOS version of this should be available from you local GNU archive
#  in the gnu/gnuish directory.
#
#  Note that some large MIBS may cause mib2html to run out of memory on
#  DOS machines.
#

function sort(array,   f, i, len, t)
{
  len = 0
  for (i in array) {
    _sort[len] = i;
    len++;
  }

  f = 1;
  while (f != 0) {
    f = 0;
    for (i = 0; i < len-1; i++) {
      if (array[_sort[i]] > array[_sort[i+1]]) {
        t = _sort[i];
        _sort[i] = _sort[i+1]
        _sort[i+1] = t;
        f = 1;
      }
    }
  }
  return len;
}

function sortobj(array,   f, i, len, t, a1, a2, l1, l2, c, j)
{
  len = 0
  for (i in array) {
    _sort[len] = i;
    len++;
  }

  f = 1;
  while (f != 0) {
    f = 0;
    for (i = 0; i < len-1; i++) {
      l1 = split(array[_sort[i]],   a1, ".")
      l2 = split(array[_sort[i+1]], a2, ".")
      c = l1 - l2;
      j = 1;
      while (c == 0 && j <= l1) {
        c = a1[j] - a2[j]
        j++
      }
      if (c > 0) {
        t = _sort[i];
        _sort[i] = _sort[i+1]
        _sort[i+1] = t;
        f = 1;
      }
    }
  }
  return len;
}

function make_filename(fn,  nfn, i)
{
  if (FORCE8DOT3) 
    nfn = substr(tolower(fn), 1, 8)
  else
    nfn = tolower(fn)
  fn  = nfn
    
  nfn = fn fileext
  if (! (nfn in filenames)) {
    filenames[nfn] = 1;
    return nfn
  }

  if (FORCE8DOT3) 
    nfn = substr(tolower(fn), 1, 6)
  else
    nfn = tolower(fn)
  fn = nfn;

  i = 1;
  nfn = fn i fileext
  while (nfn in filenames) {
    i++;
    nfn = fn i fileext
  }

  filenames[nfn] = 1;

  return nfn;
}

function output_hdr(n)
{
  print "<HTML><TITLE>" n " for <!--#snmp host--></TITLE>" > outfile
  print "<!--Standard_MIBMaster_Header-->" >> outfile
  print "<BODY><H1>" n "</H1>" >> outfile
  print "<TABLE BORDER><TR><TH>OID<TH>Name<TH>Type<TH>Value<TH>Access" >> outfile
}

function output_trailer(desc)
{
  print "</TABLE>" >> outfile
  print "<HR>" >> outfile
  print desc >> outfile
  print "</BODY></HTML>" >> outfile
  close(outfile);
}

function output_entry(group, oid, tflag)
{
  _output_entry(group, oid, "smartvalue", tflag)
}

function output_enum(group, oid, enumtext, tflag)
{
  _output_entry(group, oid, "array " enumtext, tflag)
}

function _output_entry(group, oid, text, tflag,     name, access)
{
  name   = oidToSymbol[oid]
  access = oidToAccess[oid]

  print "<TR>" >> outfile

  if (tflag != "")
    print "<TD><!--#snmp get " oid " num--><!--#snmp id-->" >> outfile
  else
    print "<TD><!--#snmp get " oid ".0--><!--#snmp id-->" >> outfile

  print "<TD><A HREF=\"#" name "\">" name "</A>" >> outfile
  print "<TD><!--#snmp type-->" >> outfile

  print "<TD><!--#snmp " text "-->" >> outfile

  if (access == "ro")
    print "<TD>Read-Only" >> outfile
  else
    print" <TD><A HREF=\"/change.html?<!--#snmp host_community-->&oid=<!--#snmp id-->\">Read-Write</A>" >> outfile
}

function output_tableref(oid, dir,     fn, name)
{
  name = oidToSymbol[oid]
  fn = make_filename(name)

  print "<TR>" >> outfile
  print "<TD valign=top>" oid ".1" >> outfile
  print "<TD valign=top><A HREF=\"#" name "\">" name "</A>" >> outfile
  print "<TD valign=top>Table" >> outfile
  print "<TD><!--#snmp table " dir fn " " oid ".1-->" >> outfile
  print "<TD>" >> outfile
  return fn
}

function make_description(name, text,    str)
{
  str = "<A NAME=\"" name "\">\n"
  str = str "<H3>" name "</H3>\n"
  str = str text "\n"
  str = str "<HR>\n"
  return str;
}

function get_new_line(  stat, line)
{
  do {
    stat = getline line;
    if (stat < 0)
      return EOF;
    gsub(/^[ \t]*/, "", line);
    gsub(/[ \t]$/,  "", line);
  } while (line == "")
  return line;
}

function get_raw_token(   n, s, t, u, z)
{
  gsub(/^[ \t]*/, "", _text);
  gsub(/[ \t]$/,  "", line);

  while (1) {
    if (_text == "")
      _text = get_new_line()
    if (_text !~ /^\-\-/) 
      break;
    _text = get_new_line();
  }

#print "_text = " _text

  if (_text ~ /^\"[^\"]*/) {
    gsub(/^\"/, "", _text);
    sval = "";
    while (1) {
      gsub(/^[ \t]*/, "", _text);
      gsub(/[ \t]*$/, "", _text);
      if (_text ~ /\"[ \t]*$/) 
        break;
      if (sval == "")
        sval = _text "\n"
      else
        sval = sval " " _text "\n"
      stat = getline _text;
      if (stat < 0)
        break;
      if (_text ~ /^[ \t]*$/)
        _text = "<P>\n"
    }
    gsub(/\"$/, "\n", _text);
    sval = sval _text
    _text = "";
    return LITERAL;
  }

  if (match(_text, /^[a-zA-Z][a-zA-Z0-9\-_]+/) == 1) {
    sval  = substr(_text, 1, RLENGTH);
    u     = substr(_text, RLENGTH+1);
    _text = u;
    return IDENTIFIER;
  }

  if (match(_text, /[0-9]+/) == 1) {
    sval  = substr(_text, 1, RLENGTH);
    u     = substr(_text, RLENGTH+1);
    _text = u;
    return INTEGER;
  }

  if (match(_text, /^\:\:\=/) == 1) {
    sval  = substr(_text, 1, RLENGTH);
    u     = substr(_text, RLENGTH+1);
    _text = u;
    return ASSIGNMENT;
  }

  t     = substr(_text, 1, 1);
  sval  = t;
  _text = substr(_text, 2);

  return t;
}

function ignore_line()
{
  _text = "";
}

function get_token(   t)
{
  if (_unget_sp > 0) {
    t    = _unget_token_stack [_unget_sp];
    sval = _unget_lexeme_stack[_unget_sp];
    _unget_sp--;
    return t;
  }

  t = get_raw_token();
  if (t == EOF)
    return t;
  if (t != IDENTIFIER)
    return t;

  if (sval in keyword) 
    return keyword[sval];

  return t;
}

function unget_token(t, s)
{
  _unget_sp++;
  _unget_token_stack [_unget_sp] = t;
  _unget_lexeme_stack[_unget_sp] = s;
}


function expect_identifier(id,   t)
{
  t = get_token();
  if (t != IDENTIFIER || sval != id) {
    print "error 1 (" NR ") : expected " id ", found " s;
    exit;
  }
}

function expect_token(t,   u)
{
  u = get_token();
  if (t != u) {
    print "error 2 (" NR ") : expected token " t " found " u; 
    exit;
  }
}

function get_integer(   t)
{
  t = get_token();
  if (t != INTEGER) {
    print "error 3 (" NR ") : expected integer, found " sval;
    exit;
  }
  return sval;
}

function get_identifier(   t)
{
  t = get_token();
  if (t != IDENTIFIER) {
    print "error 4 (" NR ") : expected identifier " s  ", found " t;
    exit;
  }
  return sval;
}

function parse_enumeratedtype(   text, t, pos, num, s)
{
  text = "";
  pos = 1;
  while (1) {
    t = get_token();
    if (t == "}")
      return text
    else if (t == ",")
      ;
    else if (t == IDENTIFIER) {
      s = sval;
      expect_token("(")
      num = get_integer()
      expect_token(")")
      while (pos < num) {
        text = text "ILLEGAL "
        pos++;
      }
      text = text s " "
      pos++;
    }
  }
}

function parse_size(  t, brackets)
{
  brackets = 0;
  while (1) {
    t = get_token()
    if (t == ")") {
      brackets--;
      if (brackets <= 0)
        return;
    } else if (t == "(")
      brackets++;
  }
}

function parse_type(  s1, t2, t)
{
  t = get_token();
  if (t == INTEGERt) {
    t = get_token();
    if (t == "{") {
      t = "__enum" internalType;
      internalType++;
      type[t] = parse_enumeratedtype();
      return t;
    } else if (t == "(") 
      ignore_line();
    else 
      unget_token(t, sval);
    return INTEGERt;
  } 

  if (t == SEQUENCE) {
    t = get_token();
    if (sval == "OF") {
      t = "__sequence" internalType;
      internalType++;
      type[t] = get_identifier();
      return t;
    } else if (t == "{") {
      s = "";
      while (1) {
        t = get_token();
        while (t == ",") 
          t = get_token()
        if (t == "}" || t == EOF)
          break;
        if (s == "")
          s = sval
        else
          s = s " " sval
        parse_type()
      }
      t = "__sequence" internalType;
      internalType++;
      type[t] = s;
      return t;
    } 
    print "error (" NR ") : malformed sequence defn"
    return 0;
  }

  if (t == OBJECT_TYPE) {
    return OBJECT_TYPE;
  }

  if (t == TRAP_TYPE) {
    return TRAP_TYPE;
  }

  if (t == OBJECT) {
    s1 = sval;
    t2 = get_token();
    if (t2 == IDENTIFIER)
      return OBJECTID;
    unget_token(t2, sval);
    unget_token(t,  s1);
  }

  if (t == OCTET) {
    s1 = sval;
    t2 = get_token();
    if (t2 == STRING) {
      t = get_token();
      if (t == "(") 
        ignore_line()
      else
        unget_token(t, sval)
      return OCTETSTRING;
    }
    unget_token(t2, sval);
    unget_token(t,  s1);
  }

  if (t == IDENTIFIER && (sval in type)) {
    s1 = sval
    t = get_token();
    if (t == "(") 
      ignore_line();
    else
      unget_token(t, sval);
    return s1;
  }

#  print "warning 1 (" NR ") : Unknown type " sval;
  return sval;
}

function parse_imports(  t)
{
  while (t != ";")
    t = get_token();
}


function parse_block(  t)
{
  expect_token("{")
  t = get_token();
  while (t != "}") {
#    print sval
    t = get_token();
  }
}

function add_to_root_list(parent,      t, oid, arr)
{
  for (t in rootList) {
    if (index(parent, t ".") == 1) 
      return;
    if (index(t, parent ".") == 1) {
      print "deleting old root " t;
      delete rootList[t];
    }
  }
  rootList[parent] = parent;
}

function parse_variables(  s)
{
  s = ""
  expect_token("{")
  while (1) {
    t = get_token()
    while (t == ",")
      t = get_token();
    if (t == "}")
      break;
    if (s == "")
      s = sval
    else
      s = s " " sval
  }
  return s
}

function parse_instancedef(q, tab, lastGroup,    n, t, t2, s, i, y, z, str, type, access, syntax, fn, array, parent)
{
  n = get_identifier();
#print "starting body for " n
  type = parse_type()
  access = "ro"
  syntax = -1;

  t = get_token();
  while (1) {
    if (t == ASSIGNMENT) 
      break;
    else if (t == SYNTAX) {
      syntax = parse_type();
    } else if (t == ACCESS) {
      t = get_token();
      if (tolower(sval) == "read-only")
        access = "ro"
      else
        access = "rw"
    } else if (t == ENTERPRISE) {
      ignore_line();
    } else if (t == STATUS) {
      ignore_line();
    } else if (t == DEFVAL) {
      ignore_line();
    } else if (t == INDEX) {
      parse_block();
    } else if (t == REFERENCE) {
      if (get_token() != LITERAL)
        ;
    } else if (t == VARIABLES) {
      variables[n] = parse_variables()
    } else if (t == DESCRIPTION) {
      if (get_token() != LITERAL)
        ;
      if (!q)
        sdesc = make_description(n, sval)
    } else {
      print "warning 2 (" NR ") :	unexpected token " sval;
    }
    t = get_token();
  }

  t = get_token()
  if (t == "{") {
    str    = "";
    parent = "";
    while (1) {
      t = get_token()
      if (t == "}")
        break;
      if (t == INTEGER) {
        parent = str
        if (str == "")
          str = sval + 0;
        else
          str = str "." sval+0;
      } else if (t == IDENTIFIER) {
        s1 = sval
        t2 = get_token()
        if (t2 == "(") {
          z = get_integer();
          expect_token(")")
          parent = str
          if (str == "")
            str = z
          else
            str = str "." z
          if (!(s1 in symbolToOid)) {
            symbolToOid[s1] = str;
            oidToSymbol[s]  = s1;
          }
        } else {
          unget_token(t2, sval)
          if (s1 in symbolToOid)
            z = symbolToOid[s1]
          else {
            print "error (" NR ") : undefined symbol " s1
            z = 0;
          }
          parent = str
          if (str == "")
            str = z
          else
            str = str "." z 
        }
      } else
        break;
    }

    if (type == OBJECT_TYPE)
      add_to_root_list(parent)
    oidParent[str] = parent;

  } else if (t == INTEGER) {
    i = sval+0;
    str = i;
  } else {
    print "error (" NR ") : unexpected token " sval;
    return 0;
  }

  symbolToOid[n]    = str;
  oidToSymbol[str]  = n;
  oidToDesc[str]    = sdesc;
  oidToAccess[str]  = access;
  oidToSyntax[str]  = syntax;
  oidToType[str]    = type;
  oidUser[str]      = 1;

  return str;
}

function parse_typedef(    n, t, s)
{
  n = get_identifier();
  expect_token(ASSIGNMENT);
  t = parse_type()
  type[n] = t;
}

function parse_definition(     defname, t, t2, t3, s, lastGroup, dir)
{
  defname = get_identifier();
  expect_token(DEFINITIONS);
  expect_token(ASSIGNMENT);
  expect_token(BEGINt);
  lastGroup = "";

  print "Parsing " defname "...";

  t = get_token();
  while (t != ENDt) {
    if (t == IMPORTS) 
      parse_imports();

    else if (t == IDENTIFIER) {
      s = sval;
      t2 = get_token();
      unget_token(t2, sval);
      unget_token(t, s);
      if (t2 == ASSIGNMENT) 
        parse_typedef();
      else 
        parse_instancedef(0, 0, lastGroup);
    } else {
      print "error 5 (" NR "): unexpected token " t;
      ignore_line();
    }
    t = get_token();
  }

  return defname
}

function globals()
{
  # globals
  _text     = "";
  _unget_sp = 0;
  lastGroup = "";
  internalType = 1;

  # token definitions
  EOF        = 0;

  i = 256;
  IDENTIFIER = i++;
  LITERAL    = i++;
  ASSIGNMENT = i++;
  INTEGER    = i++;

  IMPORTS      = i++;
  ENDt         = i++;
  BEGINt       = i++;
  OBJECT       = i++;
  IDENTIFIERt  = i++;
  OBJECT_TYPE  = i++;
  SEQUENCE     = i++;
  SYNTAX       = i++;
  ACCESS       = i++;
  STATUS       = i++;
  DEFVAL       = i++;
  DESCRIPTION  = i++;
  VARIABLES    = i++;
  INDEX        = i++;
  INTEGERt     = i++;
  OCTET        = i++;
  STRING       = i++;
  OCTECTSTRING = i++;
  OBJECTID     = i++;
  REFERENCE    = i++;
  TRAP_TYPE    = i++;
  ENTERPRISE   = i++;

  # lexeme lookup table
  keyword["IMPORTS"]     = IMPORTS;
  keyword["END"]         = ENDt;
  keyword["DEFINITIONS"] = DEFINITIONS;
  keyword["BEGIN"]       = BEGINt;
  keyword["OBJECT"]      = OBJECT;
  keyword["IDENTIFIER"]  = IDENTIFIER;
  keyword["OBJECT-TYPE"] = OBJECT_TYPE
  keyword["SEQUENCE"]    = SEQUENCE;
  keyword["SYNTAX"]      = SYNTAX;
  keyword["ACCESS"]      = ACCESS;
  keyword["STATUS"]      = STATUS;
  keyword["DEFVAL"]      = STATUS;
  keyword["DESCRIPTION"] = DESCRIPTION;
  keyword["VARIABLES"]   = VARIABLES;
  keyword["INDEX"]       = INDEX;
  keyword["INTEGER"]     = INTEGERt;
  keyword["OCTET"]       = OCTET;
  keyword["STRING"]      = STRING;
  keyword["REFERENCE"]   = REFERENCE;
  keyword["TRAP-TYPE"]   = TRAP_TYPE;
  keyword["ENTERPRISE"]  = ENTERPRISE;

  # RFC 1155 symbols
  symbolToOid["iso"]          = "1";
  symbolToOid["org"]          = "1.3";
  symbolToOid["dod"]          = "1.3.6";
  symbolToOid["internet"]     = "1.3.6.1";
  symbolToOid["directory"]    = symbolToOid["internet"] ".1";
  symbolToOid["mgmt"]         = symbolToOid["internet"] ".2";
  symbolToOid["experimental"] = symbolToOid["internet"] ".3";
  symbolToOid["private"]      = symbolToOid["internet"] ".4";
  symbolToOid["mib"]          = symbolToOid["mgmt"] ".1"
  symbolToOid["mib-2"]        = symbolToOid["mgmt"] ".1"
  symbolToOid["system"]       = symbolToOid["mib-2"] ".1"
  symbolToOid["interfaces"]   = symbolToOid["mib-2"] ".2"
  symbolToOid["at"]           = symbolToOid["mib-2"] ".3"
  symbolToOid["ip"]           = symbolToOid["mib-2"] ".4"
  symbolToOid["icmp"]         = symbolToOid["mib-2"] ".5"
  symbolToOid["tcp"]          = symbolToOid["mib-2"] ".6"
  symbolToOid["udp"]          = symbolToOid["mib-2"] ".7"
  symbolToOid["egp"]          = symbolToOid["mib-2"] ".8"
  symbolToOid["cmot"]         = symbolToOid["mib-2"] ".9"
  symbolToOid["transmission"] = symbolToOid["mib-2"] ".10"
  symbolToOid["snmp"]         = symbolToOid["mib-2"] ".11"
  symbolToOid["enterprises"]  = symbolToOid["private"] ".1"

  # types
  type["DisplayString"] = 0;
  type["Counter"] = 0;
  type["TimeTicks"] = 0;
  type["Gauge"] = 0;
  type["PhysAddress"] = 0;
  type["IpAddress"] = 0;
  type["DateAndTime"] = 0;

  for (symbol in symbolToOid) 
    oidToSymbol[symbolToOid[symbol]] = symbol
}

BEGIN {
  FORCE8DOT3 = 0;
  
  globals();


  if (FORCE8DOT3)
    fileext = ".htm"
  else
    fileext = ".html";

  topLevelFileCount = 0;

  defname = parse_definition();

  fn = tolower(defname);
  if (FORCE8DOT3) {
    p = index(fn, "-");
    if (p > 0)
      fn = substr(fn, 1, p-1);
    fn = substr(fn, 1, 8);
  }
  outfile = make_filename(fn);

  dir = tolower(defname);
  if (FORCE8DOT3) {
    p = index(dir, "-");
    if (p > 0)
      dir = substr(dir, 1, p-1);
    dir = substr(dir, 1, 8);
  }
  system("mkdir " dir)

  dir = dir "/";

  jlen = sortobj(rootList)
  for (j = 0; j < jlen; j++) 
    nrootList[j] = rootList[_sort[j]]

  if (jlen == 1) {
    idx = nrootList[0]
    print "single tree " idx
    output_topfile(dir, dir, idx, oidToSymbol[idx])
  } else {
    print "multiple trees"

    print "Creating " outfile;
    print "<HTML><TITLE>" defname " for <!--#snmp host--></TITLE>" > outfile
    print "<!--Standard_MIBMaster_Header-->" >> outfile
    print "<BODY><H1>" defname "</H1>" >> outfile
    print "<UL><H3>" >> outfile

    for (j = 0; j < jlen; j++) {
      idx    = nrootList[j]
      symbol = oidToSymbol[idx]
      fn     = make_filename(symbol)
      print "Tree " j " " idx " " symbol
      soutfile = outfile
      outfile = dir fn

      s = output_topfile("", dir, idx, symbol)

      outfile = soutfile
      if (s) {
        printf "<LI><H3><A HREF=\"" dir fn >> outfile
        printf "?<!--#snmp host_community-->\">" >> outfile
        print symbol "</A></H3>" >> outfile
      }
    }

    print "</UL><BODY></HTML>" >> outfile
  }

  close(outfile)
}

function output_topfile(ndir, dir, root, symbol,    i, hasObjectType, oid, suffix, list, t, l, nlist)
{
  i = 0
  hasObjectType = 0;
  for (oid in oidParent) {
    if (oidParent[oid] == root) {
      suffix = substr(oid, length(root)+2)
      list[i] = suffix + 0;
      t = oidToType[oid];
      if (t != OBJECTID)
        hasObjectType = 1;
      i++;
    }
  }
  if (!hasObjectType)
    return 0;

  l = sortobj(list)

  for (i = 0; i < l; i++) 
    nlist[i] = list[_sort[i]]

  return output_subfile(ndir, dir, root, symbol, nlist, i, 0) 
}

function output_table(root, symbol, defn,    i, hasObjectType, oid, suffix, list, t, sym, array, l)
{
  l = split(defn, array, " ")
  hasObjectType = 0;
  for (i = 1; i <= l; i++) {
    sym = array[i]
    oid = symbolToOid[sym]
    suffix = substr(oid, length(root)+2)
    list[i-1] = suffix;
    t = oidToType[oid];
    if (t != OBJECTID)
      hasObjectType = 1;
  }
  if (!hasObjectType)
    return 0;

  return output_subfile("", "", root, symbol, list, l, 1) 
}

function output_subfile(ndir, dir, root, symbol, list, len, isTable,      i, oid, t, suffix, t2, desc, array, l, u, sfn, soutfile, s, tflag, s2, s3)
{
  output_hdr(symbol)

  for (i = 0; i < len; i++) {
    list_i = list[i]
    oid = root "." list_i
    t = oidToType[oid]
    s = oidToSyntax[oid]
    if (t == OBJECT_TYPE) {
      if (isTable) 
        tflag = list_i;
      else
        tflag = ""
      desc = desc oidToDesc[oid]
      if (s ~ /^__enum/) 
        output_enum(symbol, oid, type[s], tflag)
      else {
        s2 = s
        s3 = ""
        while (1) {
          if (s2 == "" || !(s2 in type))
            break;
          s3 = s2;
          s2 = type[s2];
        }
        if (s3 ~ /^__sequence/) {
          sfn = output_tableref(oid, ndir)
          soutfile = outfile
          outfile  = dir sfn
          output_table(oid, symbol, s2)
          outfile  = soutfile
        } else if (s3 == "DateAndTime") 
          _output_entry(group, oid, "dateandtime", tflag)
        else
          output_entry(symbol, oid, tflag)
      }
    }
  }

  output_trailer(desc)

  return 1;
}
